# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.executors.utils.base_config_keys import BaseConfigKeys

if TYPE_CHECKING:
    from airflow.models.taskinstancekey import TaskInstanceKey


CONFIG_GROUP_NAME = "aws_lambda_executor"
INVALID_CREDENTIALS_EXCEPTIONS = [
    "ExpiredTokenException",
    "InvalidClientTokenId",
    "UnrecognizedClientException",
]


@dataclass
class LambdaQueuedTask:
    """Represents a Lambda task that is queued. The task will be run in the next heartbeat."""

    key: TaskInstanceKey
    command: CommandType
    queue: str
    executor_config: ExecutorConfigType
    attempt_number: int
    next_attempt_time: datetime.datetime


class InvokeLambdaKwargsConfigKeys(BaseConfigKeys):
    """Config keys loaded which are valid lambda invoke args."""

    FUNCTION_NAME = "function_name"
    QUALIFIER = "function_qualifier"


class AllLambdaConfigKeys(InvokeLambdaKwargsConfigKeys):
    """All config keys which are related to the Lambda Executor."""

    AWS_CONN_ID = "conn_id"
    CHECK_HEALTH_ON_STARTUP = "check_health_on_startup"
    MAX_INVOKE_ATTEMPTS = "max_invoke_attempts"
    REGION_NAME = "region_name"
    QUEUE_URL = "queue_url"
    DLQ_URL = "dead_letter_queue_url"
    END_WAIT_TIMEOUT = "end_wait_timeout"


CommandType = Sequence[str]
ExecutorConfigType = dict[str, Any]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.amazon.aws.utils.mixins import aws_template_fields


def test_aws_template_fields():
    result = aws_template_fields()
    assert result == ("aws_conn_id", "region_name", "verify")
    assert aws_template_fields() is result, "aws_template_fields expect to return cached result"

    assert aws_template_fields("foo", "aws_conn_id", "bar") == (
        "aws_conn_id",
        "bar",
        "foo",
        "region_name",
        "verify",
    )


def test_aws_template_fields_incorrect_types():
    with pytest.raises(TypeError, match="Expected that all provided arguments are strings"):
        aws_template_fields(1, None, "test")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Transfers data from Exasol database into a S3 Bucket."""

from __future__ import annotations

from unittest import mock

from airflow.providers.amazon.aws.transfers.exasol_to_s3 import ExasolToS3Operator

BASE_PATH = "airflow.providers.amazon.aws.transfers.exasol_to_s3.{}"


class TestExasolToS3Operator:
    @mock.patch(BASE_PATH.format("NamedTemporaryFile"))
    @mock.patch(BASE_PATH.format("ExasolHook"))
    @mock.patch(BASE_PATH.format("S3Hook"))
    def test_execute(self, mock_s3_hook, mock_exasol_hook, mock_local_tmp_file):
        mock_fh = mock_local_tmp_file.return_value.__enter__.return_value

        op = ExasolToS3Operator(
            task_id="task_id",
            query_or_table="query_or_table",
            key="key",
            bucket_name="bucket_name",
            replace=False,
            encrypt=True,
            gzip=False,
            acl_policy="acl_policy",
            query_params={"query_params": "1"},
            export_params={"export_params": "2"},
            exasol_conn_id="exasol_conn_id",
            aws_conn_id="aws_conn_id",
        )
        op.execute({})

        mock_s3_hook.assert_called_once_with(aws_conn_id="aws_conn_id")
        mock_exasol_hook.assert_called_once_with(exasol_conn_id="exasol_conn_id")

        mock_exasol_hook.return_value.export_to_file.assert_called_once_with(
            filename=mock_fh.name,
            query_or_table="query_or_table",
            query_params={"query_params": "1"},
            export_params={"export_params": "2"},
        )

        mock_fh.flush.assert_called_once_with()

        mock_s3_hook.return_value.load_file(
            key="key",
            bucket_name="bucket_name",
            replace=False,
            encrypt=True,
            gzip=False,
            acl_policy="acl_policy",
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.notifications.ses import SesNotifier, send_ses_notification
from airflow.providers.amazon.version_compat import NOTSET

TEST_EMAIL_PARAMS = {
    "mail_from": "from@test.com",
    "to": "to@test.com",
    "subject": "Test Subject",
    "html_content": "<p>Test Content</p>",
}

# The hook sets these default values if they are not provided
HOOK_DEFAULTS = {
    "mime_charset": "utf-8",
    "mime_subtype": "mixed",
}


class TestSesNotifier:
    def test_class_and_notifier_are_same(self):
        assert send_ses_notification is SesNotifier

    @pytest.mark.parametrize(
        "aws_conn_id",
        [
            pytest.param("aws_test_conn_id", id="custom-conn"),
            pytest.param(None, id="none-conn"),
            pytest.param(NOTSET, id="default-value"),
        ],
    )
    @pytest.mark.parametrize(
        "region_name",
        [
            pytest.param("eu-west-2", id="custom-region"),
            pytest.param(None, id="no-region"),
            pytest.param(NOTSET, id="default-value"),
        ],
    )
    def test_parameters_propagate_to_hook(self, aws_conn_id, region_name):
        """Test notifier attributes propagate to SesHook."""
        notifier_kwargs = {}
        if aws_conn_id is not NOTSET:
            notifier_kwargs["aws_conn_id"] = aws_conn_id
        if region_name is not NOTSET:
            notifier_kwargs["region_name"] = region_name

        notifier = SesNotifier(**notifier_kwargs, **TEST_EMAIL_PARAMS)
        with mock.patch("airflow.providers.amazon.aws.notifications.ses.SesHook") as mock_hook:
            hook = notifier.hook
            assert hook is notifier.hook, "Hook property not cached"
            mock_hook.assert_called_once_with(
                aws_conn_id=(aws_conn_id if aws_conn_id is not NOTSET else "aws_default"),
                region_name=(region_name if region_name is not NOTSET else None),
            )

            # Basic check for notifier
            notifier.notify({})
            mock_hook.return_value.send_email.assert_called_once_with(**TEST_EMAIL_PARAMS, **HOOK_DEFAULTS)

    @pytest.mark.asyncio
    async def test_async_notify(self):
        """Test async notification sends correctly."""
        notifier = SesNotifier(**TEST_EMAIL_PARAMS)
        with mock.patch("airflow.providers.amazon.aws.notifications.ses.SesHook") as mock_hook:
            mock_hook.return_value.asend_email = mock.AsyncMock()

            await notifier.async_notify({})

            mock_hook.return_value.asend_email.assert_called_once_with(**TEST_EMAIL_PARAMS, **HOOK_DEFAULTS)

    def test_ses_notifier_with_optional_params(self):
        """Test notifier handles all optional parameters correctly."""
        email_params = {
            **TEST_EMAIL_PARAMS,
            "files": ["test.txt"],
            "cc": ["cc@test.com"],
            "bcc": ["bcc@test.com"],
            "mime_subtype": "alternative",
            "mime_charset": "ascii",
            "reply_to": "reply@test.com",
            "return_path": "bounce@test.com",
            "custom_headers": {"X-Custom": "value"},
        }

        notifier = SesNotifier(**email_params)
        with mock.patch("airflow.providers.amazon.aws.notifications.ses.SesHook") as mock_hook:
            notifier.notify({})

            mock_hook.return_value.send_email.assert_called_once_with(**email_params)

    def test_ses_notifier_templated(self, create_dag_without_db):
        """Test template fields are properly rendered."""
        templated_params = {
            "aws_conn_id": "{{ dag.dag_id }}",
            "region_name": "{{ var_region }}",
            "mail_from": "{{ var_from }}",
            "to": "{{ var_to }}",
            "subject": "{{ var_subject }}",
            "html_content": "Hello {{ var_name }}",
            "cc": ["cc@{{ var_domain }}"],
            "bcc": ["bcc@{{ var_domain }}"],
            "reply_to": "reply@{{ var_domain }}",
        }

        notifier = SesNotifier(**templated_params)
        with mock.patch("airflow.providers.amazon.aws.notifications.ses.SesHook") as mock_hook:
            context = {
                "dag": create_dag_without_db("test_ses_notifier_templated"),
                "var_region": "us-west-1",
                "var_from": "from@example.com",
                "var_to": "to@example.com",
                "var_subject": "Test Email",
                "var_name": "John",
                "var_domain": "example.com",
            }
            notifier(context)

            mock_hook.assert_called_once_with(
                aws_conn_id="test_ses_notifier_templated",
                region_name="us-west-1",
            )
            mock_hook.return_value.send_email.assert_called_once_with(
                mail_from="from@example.com",
                to="to@example.com",
                subject="Test Email",
                html_content="Hello John",
                cc=["cc@example.com"],
                bcc=["bcc@example.com"],
                mime_subtype="mixed",
                mime_charset="utf-8",
                reply_to="reply@example.com",
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import inspect
import re
import sys
from unittest import mock

import boto3
import botocore
import pytest

from airflow.providers.amazon.aws.hooks.bedrock import BedrockHook
from airflow.providers.amazon.aws.sensors.bedrock import (
    BedrockBatchInferenceSensor,
    BedrockCustomizeModelCompletedSensor,
    BedrockProvisionModelThroughputCompletedSensor,
)


class TestBedrockCustomWaiters:
    def test_service_waiters(self):
        """Ensure that all custom Bedrock waiters have unit tests."""

        def _class_tests_a_waiter(class_name: str) -> bool:
            """Check if the class name starts with 'Test' and ends with 'Waiter'."""
            return bool(re.match(r"^Test[A-Za-z]+Waiter$", class_name))

        # Collect WAITER_NAME from each waiter test class in this module.
        test_classes = inspect.getmembers(sys.modules[__name__], inspect.isclass)
        waiters_tested = [cls.WAITER_NAME for (name, cls) in test_classes if _class_tests_a_waiter(name)]

        assert sorted(BedrockHook()._list_custom_waiters()) == sorted(waiters_tested)


class TestBedrockCustomWaitersBase:
    @pytest.fixture(autouse=True)
    def mock_conn(self, monkeypatch):
        self.client = boto3.client("bedrock")
        monkeypatch.setattr(BedrockHook, "conn", self.client)


class TestModelCustomizationJobCompleteWaiter(TestBedrockCustomWaitersBase):
    WAITER_NAME = "model_customization_job_complete"

    @pytest.fixture
    def mock_get_job(self):
        with mock.patch.object(self.client, "get_model_customization_job") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", BedrockCustomizeModelCompletedSensor.SUCCESS_STATES)
    def test_model_customization_job_complete(self, state, mock_get_job):
        mock_get_job.return_value = {"status": state}

        BedrockHook().get_waiter(self.WAITER_NAME).wait(jobIdentifier="job_id")

    @pytest.mark.parametrize("state", BedrockCustomizeModelCompletedSensor.FAILURE_STATES)
    def test_model_customization_job_failed(self, state, mock_get_job):
        mock_get_job.return_value = {"status": state}

        with pytest.raises(botocore.exceptions.WaiterError):
            BedrockHook().get_waiter(self.WAITER_NAME).wait(jobIdentifier="job_id")

    def test_model_customization_job_wait(self, mock_get_job):
        wait = {"status": "InProgress"}
        success = {"status": "Completed"}
        mock_get_job.side_effect = [wait, wait, success]

        BedrockHook().get_waiter(self.WAITER_NAME).wait(
            jobIdentifier="job_id", WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )


class TestProvisionedModelThroughputCompleteWaiter(TestBedrockCustomWaitersBase):
    WAITER_NAME = "provisioned_model_throughput_complete"

    @pytest.fixture
    def mock_get_job(self):
        with mock.patch.object(self.client, "get_provisioned_model_throughput") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", BedrockProvisionModelThroughputCompletedSensor.SUCCESS_STATES)
    def test_model_customization_job_complete(self, state, mock_get_job):
        mock_get_job.return_value = {"status": state}

        BedrockHook().get_waiter(self.WAITER_NAME).wait(jobIdentifier="job_id")

    @pytest.mark.parametrize("state", BedrockProvisionModelThroughputCompletedSensor.FAILURE_STATES)
    def test_model_customization_job_failed(self, state, mock_get_job):
        mock_get_job.return_value = {"status": state}

        with pytest.raises(botocore.exceptions.WaiterError):
            BedrockHook().get_waiter(self.WAITER_NAME).wait(jobIdentifier="job_id")

    def test_model_customization_job_wait(self, mock_get_job):
        wait = {"status": "Creating"}
        success = {"status": "InService"}
        mock_get_job.side_effect = [wait, wait, success]

        BedrockHook().get_waiter(self.WAITER_NAME).wait(
            jobIdentifier="job_id", WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )


class TestBatchInferenceCompleteWaiter(TestBedrockCustomWaitersBase):
    WAITER_NAME = "batch_inference_complete"
    SENSOR = BedrockBatchInferenceSensor(
        task_id="task_id",
        job_arn="job_arn",
        success_state=BedrockBatchInferenceSensor.SuccessState.COMPLETED,
    )

    @pytest.fixture
    def mock_get_job(self):
        with mock.patch.object(self.client, "get_model_invocation_job") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", SENSOR.SUCCESS_STATES)
    def test_batch_inference_complete(self, state, mock_get_job):
        mock_get_job.return_value = {"status": state}

        BedrockHook().get_waiter(self.WAITER_NAME).wait(jobIdentifier="job_arn")

    @pytest.mark.parametrize("state", SENSOR.FAILURE_STATES)
    def test_batch_inference_failed(self, state, mock_get_job):
        mock_get_job.return_value = {"status": state}

        with pytest.raises(botocore.exceptions.WaiterError):
            BedrockHook().get_waiter(self.WAITER_NAME).wait(jobIdentifier="job_arn")

    def test_batch_inference_wait(self, mock_get_job):
        wait = {"status": "InProgress"}
        success = {"status": "Completed"}
        mock_get_job.side_effect = [wait, wait, success]

        BedrockHook().get_waiter(self.WAITER_NAME).wait(
            jobIdentifier="job_arn", WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )


class TestBatchInferenceScheduledWaiter(TestBedrockCustomWaitersBase):
    WAITER_NAME = "batch_inference_scheduled"
    SENSOR = BedrockBatchInferenceSensor(
        task_id="task_id",
        job_arn="job_arn",
        success_state=BedrockBatchInferenceSensor.SuccessState.SCHEDULED,
    )

    @pytest.fixture
    def mock_get_job(self):
        with mock.patch.object(self.client, "get_model_invocation_job") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", SENSOR.SUCCESS_STATES)
    def test_batch_inference_complete(self, state, mock_get_job):
        mock_get_job.return_value = {"status": state}

        BedrockHook().get_waiter(self.WAITER_NAME).wait(jobIdentifier="job_arn")

    @pytest.mark.parametrize("state", SENSOR.FAILURE_STATES)
    def test_batch_inference_failed(self, state, mock_get_job):
        mock_get_job.return_value = {"status": state}

        with pytest.raises(botocore.exceptions.WaiterError):
            BedrockHook().get_waiter(self.WAITER_NAME).wait(jobIdentifier="job_arn")

    def test_batch_inference_wait(self, mock_get_job):
        wait = {"status": "InProgress"}
        success = {"status": "Completed"}
        mock_get_job.side_effect = [wait, wait, success]

        BedrockHook().get_waiter(self.WAITER_NAME).wait(
            jobIdentifier="job_arn", WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from airflow.providers.amazon.aws.sensors.emr import EmrBaseSensor
from airflow.providers.common.compat.sdk import AirflowException

if TYPE_CHECKING:
    from airflow.sdk import Context

TARGET_STATE = "TARGET_STATE"
FAILED_STATE = "FAILED_STATE"
NON_TARGET_STATE = "NON_TARGET_STATE"

GOOD_HTTP_STATUS = 200
BAD_HTTP_STATUS = 400

EXPECTED_CODE = "EXPECTED_TEST_FAILURE"
EMPTY_CODE = "No code"


class EmrBaseSensorSubclass(EmrBaseSensor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_states = [TARGET_STATE]
        self.failed_states = [FAILED_STATE]
        self.response = {}  # will be set in tests

    def get_emr_response(self, context: Context):
        return self.response

    @staticmethod
    def state_from_response(response):
        return response["SomeKey"]["State"]

    @staticmethod
    def failure_message_from_response(response):
        change_reason = response["SomeKey"].get("StateChangeReason")
        if change_reason:
            return (
                f"for code: {change_reason.get('Code', EMPTY_CODE)} "
                f"with message {change_reason.get('Message', 'Unknown')}"
            )
        return None


class TestEmrBaseSensor:
    def test_poke_returns_true_when_state_is_in_target_states(self):
        operator = EmrBaseSensorSubclass(
            task_id="test_task",
            poke_interval=2,
        )
        operator.response = {
            "SomeKey": {"State": TARGET_STATE},
            "ResponseMetadata": {"HTTPStatusCode": GOOD_HTTP_STATUS},
        }

        operator.execute(None)

    def test_poke_returns_false_when_state_is_not_in_target_states(self):
        operator = EmrBaseSensorSubclass(
            task_id="test_task",
            poke_interval=2,
        )
        operator.response = {
            "SomeKey": {"State": NON_TARGET_STATE},
            "ResponseMetadata": {"HTTPStatusCode": GOOD_HTTP_STATUS},
        }

        assert operator.poke(None) is False

    def test_poke_returns_false_when_http_response_is_bad(self):
        operator = EmrBaseSensorSubclass(
            task_id="test_task",
            poke_interval=2,
        )
        operator.response = {
            "SomeKey": {"State": TARGET_STATE},
            "ResponseMetadata": {"HTTPStatusCode": BAD_HTTP_STATUS},
        }

        assert operator.poke(None) is False

    def test_poke_raises_error_when_state_is_in_failed_states(self):
        operator = EmrBaseSensorSubclass(
            task_id="test_task",
            poke_interval=2,
        )
        operator.response = {
            "SomeKey": {"State": FAILED_STATE, "StateChangeReason": {"Code": EXPECTED_CODE}},
            "ResponseMetadata": {"HTTPStatusCode": GOOD_HTTP_STATUS},
        }

        with pytest.raises(AirflowException) as ctx:
            operator.poke(None)

        assert "EMR job failed" in str(ctx.value)
        assert EXPECTED_CODE in str(ctx.value)
        assert EMPTY_CODE not in str(ctx.value)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock
from uuid import UUID

import pytest
from botocore.exceptions import WaiterError

from airflow.providers.amazon.aws.hooks.emr import EmrServerlessHook
from airflow.providers.amazon.aws.operators.emr import (
    EmrServerlessCreateApplicationOperator,
    EmrServerlessDeleteApplicationOperator,
    EmrServerlessStartJobOperator,
    EmrServerlessStopApplicationOperator,
)
from airflow.providers.amazon.version_compat import NOTSET
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

if TYPE_CHECKING:
    from unittest.mock import MagicMock

task_id = "test_emr_serverless_task_id"
application_id = "test_application_id"
release_label = "test"
job_type = "test"
client_request_token = "eac427d0-1c6d-4dfb9a-32423412"
config = {"name": "test_application_emr_serverless"}

execution_role_arn = "test_emr_serverless_role_arn"
job_driver = {"test_key": "test_value"}
spark_job_driver = {"sparkSubmit": {"entryPoint": "test.py"}}
configuration_overrides = {"monitoringConfiguration": {"test_key": "test_value"}}
job_run_id = "test_job_run_id"
s3_logs_location = "s3://test_bucket/test_key/"
cloudwatch_logs_group_name = "/aws/emrs"
cloudwatch_logs_prefix = "myapp"
s3_configuration_overrides = {
    "monitoringConfiguration": {"s3MonitoringConfiguration": {"logUri": s3_logs_location}}
}
cloudwatch_configuration_overrides = {
    "monitoringConfiguration": {
        "cloudWatchLoggingConfiguration": {
            "enabled": True,
            "logGroupName": cloudwatch_logs_group_name,
            "logStreamNamePrefix": cloudwatch_logs_prefix,
        }
    }
}

application_id_delete_operator = "test_emr_serverless_delete_application_operator"


class TestEmrServerlessCreateApplicationOperator:
    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_execute_successfully_with_wait_for_completion(self, mock_conn, mock_waiter):
        mock_waiter().wait.return_value = True
        mock_conn.create_application.return_value = {
            "applicationId": application_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        mock_conn.get_application.side_effect = [
            {"application": {"state": "CREATED"}},
            {"application": {"state": "STARTED"}},
        ]

        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            config=config,
            waiter_max_attempts=3,
            waiter_delay=0,
        )

        id = operator.execute(None)

        mock_conn.create_application.assert_called_once_with(
            clientToken=client_request_token,
            releaseLabel=release_label,
            type=job_type,
            **config,
        )
        mock_waiter().wait.assert_called_with(
            applicationId=application_id,
            WaiterConfig={
                "MaxAttempts": 1,
            },
        )
        assert mock_waiter().wait.call_count == 2

        mock_conn.start_application.assert_called_once_with(applicationId=application_id)
        assert id == application_id
        mock_conn.get_application.call_count == 2

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_execute_successfully_no_wait_for_completion(self, mock_conn, mock_waiter):
        mock_waiter().wait.return_value = True
        mock_conn.create_application.return_value = {
            "applicationId": application_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            wait_for_completion=False,
            config=config,
        )

        id = operator.execute(None)

        mock_conn.create_application.assert_called_once_with(
            clientToken=client_request_token,
            releaseLabel=release_label,
            type=job_type,
            **config,
        )
        mock_conn.start_application.assert_called_once_with(applicationId=application_id)

        mock_waiter().wait.assert_called_once()
        assert id == application_id

    @mock.patch.object(EmrServerlessHook, "conn")
    def test_failed_create_application_request(self, mock_conn):
        mock_conn.create_application.return_value = {
            "applicationId": application_id,
            "ResponseMetadata": {"HTTPStatusCode": 404},
        }

        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            config=config,
        )

        with pytest.raises(AirflowException) as ex_message:
            operator.execute(None)

        assert "Application Creation failed:" in str(ex_message.value)

        mock_conn.create_application.assert_called_once_with(
            clientToken=client_request_token,
            releaseLabel=release_label,
            type=job_type,
            **config,
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_failed_create_application(self, mock_conn, mock_get_waiter):
        error = WaiterError(
            name="test_name",
            reason="Waiter encountered a terminal failure state:",
            last_response={"application": {"state": "FAILED"}},
        )
        mock_get_waiter().wait.side_effect = error
        mock_conn.create_application.return_value = {
            "applicationId": application_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            config=config,
        )

        with pytest.raises(AirflowException) as ex_message:
            operator.execute(None)

        assert "Serverless Application creation failed:" in str(ex_message.value)

        mock_conn.create_application.assert_called_once_with(
            clientToken=client_request_token,
            releaseLabel=release_label,
            type=job_type,
            **config,
        )
        mock_conn.create_application.return_value = {
            "applicationId": application_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        error = WaiterError(
            name="test_name",
            reason="Waiter encountered a terminal failure state:",
            last_response={"application": {"state": "TERMINATED"}},
        )
        mock_get_waiter().wait.side_effect = error

        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            config=config,
        )

        with pytest.raises(AirflowException) as ex_message:
            operator.execute(None)

        assert "Serverless Application creation failed:" in str(ex_message.value)

        mock_conn.create_application.assert_called_with(
            clientToken=client_request_token,
            releaseLabel=release_label,
            type=job_type,
            **config,
        )
        mock_conn.create_application.call_count == 2

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_failed_start_application(self, mock_conn, mock_get_waiter):
        error = WaiterError(
            name="test_name",
            reason="Waiter encountered a terminal failure state:",
            last_response={"application": {"state": "TERMINATED"}},
        )
        mock_get_waiter().wait.side_effect = [True, error]
        mock_conn.create_application.return_value = {
            "applicationId": application_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            config=config,
        )

        with pytest.raises(AirflowException) as ex_message:
            operator.execute(None)

        assert "Serverless Application failed to start:" in str(ex_message.value)

        mock_conn.create_application.assert_called_once_with(
            clientToken=client_request_token,
            releaseLabel=release_label,
            type=job_type,
            **config,
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_no_client_request_token(self, mock_conn, mock_waiter):
        mock_waiter().wait.return_value = True
        mock_conn.create_application.return_value = {
            "applicationId": application_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            wait_for_completion=False,
            config=config,
        )

        operator.execute(None)
        generated_client_token = operator.client_request_token

        assert str(UUID(generated_client_token, version=4)) == generated_client_token

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_application_in_failure_state(self, mock_conn, mock_get_waiter):
        fail_state = "STOPPED"
        error = WaiterError(
            name="test_name",
            reason="Waiter encountered a terminal failure state:",
            last_response={"application": {"state": fail_state}},
        )
        mock_get_waiter().wait.side_effect = [error]
        mock_conn.create_application.return_value = {
            "applicationId": application_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            config=config,
        )

        with pytest.raises(AirflowException) as ex_message:
            operator.execute(None)

        assert str(ex_message.value) == f"Serverless Application creation failed: {error}"

        mock_conn.create_application.assert_called_once_with(
            clientToken=client_request_token,
            releaseLabel=release_label,
            type=job_type,
            **config,
        )

    @pytest.mark.parametrize(
        ("waiter_delay", "waiter_max_attempts", "expected"),
        [
            (NOTSET, NOTSET, [60, 25]),
            (30, 10, [30, 10]),
        ],
    )
    def test_create_application_waiter_params(
        self,
        waiter_delay,
        waiter_max_attempts,
        expected,
    ):
        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            config=config,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
        )
        assert operator.wait_for_completion is True
        assert operator.waiter_delay == expected[0]
        assert operator.waiter_max_attempts == expected[1]

    @mock.patch.object(EmrServerlessHook, "conn")
    def test_create_application_deferrable(self, mock_conn):
        mock_conn.create_application.return_value = {
            "applicationId": application_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            config=config,
            deferrable=True,
        )

        with pytest.raises(TaskDeferred):
            operator.execute(None)

    def test_template_fields(self):
        operator = EmrServerlessCreateApplicationOperator(
            task_id=task_id,
            release_label=release_label,
            job_type=job_type,
            client_request_token=client_request_token,
            config=config,
            waiter_max_attempts=3,
            waiter_delay=0,
        )

        template_fields = list(operator.template_fields) + list(operator.template_fields_renderers.keys())

        class_fields = operator.__dict__

        missing_fields = [field for field in template_fields if field not in class_fields]

        assert not missing_fields, f"Templated fields are not available {missing_fields}"


class TestEmrServerlessStartJobOperator:
    def setup_method(self):
        self.mock_context = mock.MagicMock()

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_job_run_app_started(self, mock_conn, mock_get_waiter):
        mock_get_waiter().wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        mock_conn.get_job_run.return_value = {"jobRun": {"state": "SUCCESS"}}

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
        )
        id = operator.execute(self.mock_context)
        default_name = operator.name

        assert operator.wait_for_completion is True
        mock_conn.get_application.assert_called_once_with(applicationId=application_id)
        assert id == job_run_id
        mock_conn.start_job_run.assert_called_once_with(
            clientToken=client_request_token,
            applicationId=application_id,
            executionRoleArn=execution_role_arn,
            jobDriver=job_driver,
            configurationOverrides=configuration_overrides,
            name=default_name,
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_job_run_job_failed(self, mock_conn, mock_get_waiter):
        error = WaiterError(
            name="test_name",
            reason="Waiter encountered a terminal failure state:",
            last_response={"jobRun": {"state": "FAILED"}},
        )
        mock_get_waiter().wait.side_effect = [error]
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
        )
        with pytest.raises(AirflowException) as ex_message:
            operator.execute(self.mock_context)

        assert "Serverless Job failed:" in str(ex_message.value)
        default_name = operator.name

        mock_conn.get_application.assert_called_once_with(applicationId=application_id)
        mock_conn.start_job_run.assert_called_once_with(
            clientToken=client_request_token,
            applicationId=application_id,
            executionRoleArn=execution_role_arn,
            jobDriver=job_driver,
            configurationOverrides=configuration_overrides,
            name=default_name,
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_job_run_app_not_started(self, mock_conn, mock_get_waiter):
        mock_get_waiter().wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "CREATING"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
        )
        id = operator.execute(self.mock_context)
        default_name = operator.name

        assert operator.wait_for_completion is True
        mock_conn.get_application.assert_called_once_with(applicationId=application_id)
        assert mock_get_waiter().wait.call_count == 2
        assert id == job_run_id
        mock_conn.start_job_run.assert_called_once_with(
            clientToken=client_request_token,
            applicationId=application_id,
            executionRoleArn=execution_role_arn,
            jobDriver=job_driver,
            configurationOverrides=configuration_overrides,
            name=default_name,
        )

    @mock.patch("time.sleep", return_value=None)
    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_execute_max_attempts_cancel_job(self, mock_conn, mock_get_waiter, sleep_mock):
        application_id = "test-app-id"
        job_run_id = "test-job-id"
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        mock_get_waiter.return_value.wait.side_effect = AirflowException("Waiter error: max attempts reached")
        operator = EmrServerlessStartJobOperator(
            task_id="test_task",
            application_id=application_id,
            execution_role_arn="arn:aws:iam::123456789012:role/test-role",
            job_driver={"sparkSubmit": {"entryPoint": "s3://my-bucket/my-script.py"}},
            client_request_token="token",
            configuration_overrides={},
            waiter_delay=1,
            waiter_max_attempts=1,
            wait_for_completion=True,
        )

        mock_context = mock.MagicMock()
        with mock.patch.object(operator.log, "info") as mock_log_info:
            with pytest.raises(AirflowException, match="Waiter error: max attempts reached"):
                operator.execute(mock_context)
            mock_conn.cancel_job_run.assert_called_once_with(
                applicationId=application_id, jobRunId=job_run_id
            )
            log_msgs = [call.args[0] for call in mock_log_info.call_args_list]
            assert any("Cancelling EMR Serverless job" in msg for msg in log_msgs)

    @mock.patch("time.sleep", return_value=None)
    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_execute_max_attempts_cancel_job_cancel_fails(self, mock_conn, mock_get_waiter, sleep_mock):
        application_id = "test-app-id"
        job_run_id = "test-job-id"
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        mock_get_waiter.return_value.wait.side_effect = AirflowException("Waiter error: max attempts reached")
        mock_conn.cancel_job_run.side_effect = Exception(
            "Failed to cancel EMR Serverless job after waiter timeout"
        )
        operator = EmrServerlessStartJobOperator(
            task_id="test_task",
            application_id=application_id,
            execution_role_arn="arn:aws:iam::123456789012:role/test-role",
            job_driver={"sparkSubmit": {"entryPoint": "s3://my-bucket/my-script.py"}},
            client_request_token="token",
            configuration_overrides={},
            waiter_delay=1,
            waiter_max_attempts=1,
            wait_for_completion=True,
        )
        mock_context = mock.MagicMock()

        with (
            mock.patch.object(operator.log, "exception") as mock_log_exception,
            mock.patch.object(operator.log, "info") as mock_log_info,
        ):
            with pytest.raises(AirflowException, match="Waiter error: max attempts reached"):
                operator.execute(mock_context)
            mock_conn.cancel_job_run.assert_called_once_with(
                applicationId=application_id,
                jobRunId=job_run_id,
            )
            log_msgs = [call.args[0] for call in mock_log_exception.call_args_list]
            assert any("Failed to cancel EMR Serverless job" in msg for msg in log_msgs)
            info_msgs = [call.args[0] for call in mock_log_info.call_args_list]
            assert any("Cancelling EMR Serverless job" in msg for msg in info_msgs)

    @mock.patch("time.sleep", return_value=None)
    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_execute_waiter_other_exception_does_not_cancel(self, mock_conn, mock_get_waiter, sleep_mock):
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": "job-id",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        mock_get_waiter.return_value.wait.side_effect = AirflowException("Other failure")

        operator = EmrServerlessStartJobOperator(
            task_id="test_task",
            application_id="app-id",
            execution_role_arn="arn",
            job_driver={"sparkSubmit": {"entryPoint": "s3://x"}},
            client_request_token="token",
            waiter_delay=1,
            waiter_max_attempts=1,
            wait_for_completion=True,
        )
        with pytest.raises(AirflowException, match="Other failure"):
            operator.execute(mock.MagicMock())
        mock_conn.cancel_job_run.assert_not_called()

    @mock.patch("time.sleep", return_value=None)
    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_execute_no_wait_for_completion_does_not_cancel(self, mock_conn, mock_get_waiter, sleep_mock):
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": "job-id",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id="test_task",
            application_id="app-id",
            execution_role_arn="arn",
            job_driver={"sparkSubmit": {"entryPoint": "s3://x"}},
            client_request_token="token",
            wait_for_completion=False,
        )

        job_id = operator.execute(mock.MagicMock())
        assert job_id == "job-id"
        mock_conn.cancel_job_run.assert_not_called()

    @mock.patch("time.sleep", return_value=None)
    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_execute_deferrable_does_not_cancel(self, mock_conn, mock_get_waiter, sleep_mock):
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": "job-id",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id="test_task",
            application_id="app-id",
            execution_role_arn="arn",
            job_driver={"sparkSubmit": {"entryPoint": "s3://x"}},
            client_request_token="token",
            wait_for_completion=True,
            deferrable=True,
        )
        operator.defer = mock.MagicMock()
        operator.execute(mock.MagicMock())
        operator.defer.assert_called_once()
        mock_conn.cancel_job_run.assert_not_called()

    @mock.patch("time.sleep", return_value=None)
    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_execute_complete_deferrable_failure_triggers_cancel(
        self, mock_conn, mock_get_waiter, sleep_mock
    ):
        application_id = "test-app-id"
        job_run_id = "test-job-id"
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        operator = EmrServerlessStartJobOperator(
            task_id="test_task",
            application_id=application_id,
            execution_role_arn="arn:aws:iam::123456789012:role/test-role",
            job_driver={"sparkSubmit": {"entryPoint": "s3://bucket/script.py"}},
            client_request_token="token",
            configuration_overrides={},
            waiter_delay=1,
            waiter_max_attempts=3,
            wait_for_completion=True,
            deferrable=True,
        )
        operator.job_id = job_run_id
        failed_event = {
            "status": "error",
            "job_details": {"application_id": application_id, "job_id": job_run_id},
        }

        mock_context = mock.MagicMock()
        with mock.patch.object(operator.log, "info") as mock_log:
            with pytest.raises(
                AirflowException, match="EMR Serverless job failed or timed out in deferrable mode"
            ):
                operator.execute_complete(mock_context, failed_event)
            mock_conn.cancel_job_run.assert_called_once_with(
                applicationId=application_id,
                jobRunId=job_run_id,
            )
            log_msgs = [call.args[0] for call in mock_log.call_args_list]
            assert any("Cancelling EMR Serverless job" in msg for msg in log_msgs)

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_job_run_app_not_started_no_wait_for_completion(self, mock_conn, mock_get_waiter):
        mock_get_waiter().wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "CREATING"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
            wait_for_completion=False,
        )
        id = operator.execute(self.mock_context)
        default_name = operator.name

        mock_conn.get_application.assert_called_once_with(applicationId=application_id)
        mock_get_waiter().wait.assert_called_once()
        assert id == job_run_id
        mock_conn.start_job_run.assert_called_once_with(
            clientToken=client_request_token,
            applicationId=application_id,
            executionRoleArn=execution_role_arn,
            jobDriver=job_driver,
            configurationOverrides=configuration_overrides,
            name=default_name,
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_job_run_app_started_no_wait_for_completion(self, mock_conn, mock_get_waiter):
        mock_get_waiter().wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
            wait_for_completion=False,
        )
        id = operator.execute(self.mock_context)
        assert id == job_run_id
        default_name = operator.name

        mock_conn.start_job_run.assert_called_once_with(
            clientToken=client_request_token,
            applicationId=application_id,
            executionRoleArn=execution_role_arn,
            jobDriver=job_driver,
            configurationOverrides=configuration_overrides,
            name=default_name,
        )
        assert not mock_get_waiter().wait.called

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_failed_start_job_run(self, mock_conn, mock_get_waiter):
        mock_get_waiter().wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "CREATING"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 404},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
        )
        with pytest.raises(AirflowException) as ex_message:
            operator.execute(self.mock_context)
        assert "EMR serverless job failed to start:" in str(ex_message.value)
        default_name = operator.name

        mock_conn.get_application.assert_called_once_with(applicationId=application_id)
        mock_get_waiter().wait.assert_called_once()
        mock_conn.start_job_run.assert_called_once_with(
            clientToken=client_request_token,
            applicationId=application_id,
            executionRoleArn=execution_role_arn,
            jobDriver=job_driver,
            configurationOverrides=configuration_overrides,
            name=default_name,
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_start_job_run_fail_on_wait_for_completion(self, mock_conn, mock_get_waiter):
        error = WaiterError(
            name="mock_waiter_error",
            reason="Waiter encountered a terminal failure state:",
            last_response={"jobRun": {"state": "FAILED", "stateDetails": "Test Details"}},
        )
        mock_get_waiter().wait.side_effect = [error]
        mock_conn.get_application.return_value = {"application": {"state": "CREATED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
        )
        with pytest.raises(AirflowException) as ex_message:
            operator.execute(self.mock_context)
        assert "Serverless Job failed:" in str(ex_message.value)
        default_name = operator.name

        mock_conn.get_application.call_count == 2
        mock_conn.start_job_run.assert_called_once_with(
            clientToken=client_request_token,
            applicationId=application_id,
            executionRoleArn=execution_role_arn,
            jobDriver=job_driver,
            configurationOverrides=configuration_overrides,
            name=default_name,
        )
        mock_get_waiter().wait.assert_called_once()

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_start_job_default_name(self, mock_conn, mock_get_waiter):
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        mock_get_waiter().wait.return_value = True

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
        )
        operator.execute(self.mock_context)
        default_name = operator.name
        generated_name_uuid = default_name.split("_")[-1]
        assert default_name.startswith("emr_serverless_job_airflow")

        mock_conn.start_job_run.assert_called_once_with(
            clientToken=client_request_token,
            applicationId=application_id,
            executionRoleArn=execution_role_arn,
            jobDriver=job_driver,
            configurationOverrides=configuration_overrides,
            name=f"emr_serverless_job_airflow_{UUID(generated_name_uuid, version=4)}",
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_start_job_custom_name(self, mock_conn, mock_get_waiter):
        mock_get_waiter().wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        custom_name = "test_name"
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
            name=custom_name,
        )
        operator.execute(self.mock_context)

        mock_conn.start_job_run.assert_called_once_with(
            clientToken=client_request_token,
            applicationId=application_id,
            executionRoleArn=execution_role_arn,
            jobDriver=job_driver,
            configurationOverrides=configuration_overrides,
            name=custom_name,
        )

    @mock.patch.object(EmrServerlessHook, "conn")
    def test_cancel_job_run(self, mock_conn):
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        mock_conn.get_job_run.return_value = {"jobRun": {"state": "RUNNING"}}

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
            wait_for_completion=False,
        )

        id = operator.execute(self.mock_context)
        operator.on_kill()
        mock_conn.cancel_job_run.assert_called_once_with(
            applicationId=application_id,
            jobRunId=id,
        )

    @pytest.mark.parametrize(
        ("waiter_delay", "waiter_max_attempts", "expected"),
        [
            (NOTSET, NOTSET, [60, 25]),
            (30, 10, [30, 10]),
        ],
    )
    def test_start_job_waiter_params(
        self,
        waiter_delay,
        waiter_max_attempts,
        expected,
    ):
        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
        )
        assert operator.wait_for_completion is True
        assert operator.waiter_delay == expected[0]
        assert operator.waiter_max_attempts == expected[1]

    @mock.patch.object(EmrServerlessHook, "conn")
    def test_start_job_deferrable(self, mock_conn):
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
            deferrable=True,
        )

        with pytest.raises(TaskDeferred):
            operator.execute(self.mock_context)

    @mock.patch.object(EmrServerlessHook, "conn")
    def test_start_job_deferrable_without_wait_for_completion(self, mock_conn):
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
            deferrable=True,
            wait_for_completion=False,
        )

        result = operator.execute(self.mock_context)

        assert result == job_run_id

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_start_job_deferrable_app_not_started(self, mock_conn, mock_get_waiter):
        mock_get_waiter.wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "CREATING"}}
        mock_conn.start_application.return_value = {
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
            deferrable=True,
        )

        with pytest.raises(TaskDeferred):
            operator.execute(self.mock_context)

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessCloudWatchLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessDashboardLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessS3LogsLink.persist")
    def test_links_start_job_default(
        self,
        mock_s3_logs_link,
        mock_logs_link,
        mock_dashboard_link,
        mock_cloudwatch_link,
        mock_conn,
        mock_get_waiter,
    ):
        mock_get_waiter.wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
        )
        operator.execute(self.mock_context)
        mock_conn.start_job_run.assert_called_once()

        mock_s3_logs_link.assert_not_called()
        mock_logs_link.assert_not_called()
        mock_dashboard_link.assert_not_called()
        mock_cloudwatch_link.assert_not_called()

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessCloudWatchLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessDashboardLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessS3LogsLink.persist")
    def test_links_s3_enabled(
        self,
        mock_s3_logs_link,
        mock_logs_link,
        mock_dashboard_link,
        mock_cloudwatch_link,
        mock_conn,
        mock_get_waiter,
    ):
        mock_get_waiter.wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=s3_configuration_overrides,
        )
        operator.execute(self.mock_context)
        mock_conn.start_job_run.assert_called_once()

        mock_logs_link.assert_not_called()
        mock_dashboard_link.assert_not_called()
        mock_cloudwatch_link.assert_not_called()
        mock_s3_logs_link.assert_called_once_with(
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            aws_partition=mock.ANY,
            log_uri=s3_logs_location,
            application_id=application_id,
            job_run_id=job_run_id,
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessCloudWatchLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessDashboardLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessS3LogsLink.persist")
    def test_links_cloudwatch_enabled(
        self,
        mock_s3_logs_link,
        mock_logs_link,
        mock_dashboard_link,
        mock_cloudwatch_link,
        mock_conn,
        mock_get_waiter,
    ):
        mock_get_waiter.wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=cloudwatch_configuration_overrides,
        )
        operator.execute(self.mock_context)
        mock_conn.start_job_run.assert_called_once()

        mock_logs_link.assert_not_called()
        mock_dashboard_link.assert_not_called()
        mock_s3_logs_link.assert_not_called()
        mock_cloudwatch_link.assert_called_once_with(
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            aws_partition=mock.ANY,
            awslogs_group=cloudwatch_logs_group_name,
            stream_prefix=f"{cloudwatch_logs_prefix}/applications/{application_id}/jobs/{job_run_id}",
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessCloudWatchLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessDashboardLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessS3LogsLink.persist")
    def test_links_applicationui_enabled(
        self,
        mock_s3_logs_link,
        mock_logs_link,
        mock_dashboard_link,
        mock_cloudwatch_link,
        mock_conn,
        mock_get_waiter,
    ):
        mock_get_waiter.wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=cloudwatch_configuration_overrides,
            enable_application_ui_links=True,
        )
        operator.execute(self.mock_context)
        mock_conn.start_job_run.assert_called_once()

        mock_logs_link.assert_not_called()
        mock_s3_logs_link.assert_not_called()
        mock_dashboard_link.assert_called_with(
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            aws_partition=mock.ANY,
            conn_id=mock.ANY,
            application_id=application_id,
            job_run_id=job_run_id,
        )
        mock_cloudwatch_link.assert_called_once_with(
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            aws_partition=mock.ANY,
            awslogs_group=cloudwatch_logs_group_name,
            stream_prefix=f"{cloudwatch_logs_prefix}/applications/{application_id}/jobs/{job_run_id}",
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessCloudWatchLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessDashboardLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessS3LogsLink.persist")
    def test_links_applicationui_with_spark_enabled(
        self,
        mock_s3_logs_link,
        mock_logs_link,
        mock_dashboard_link,
        mock_cloudwatch_link,
        mock_conn,
        mock_get_waiter,
    ):
        mock_get_waiter.wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=spark_job_driver,
            configuration_overrides=s3_configuration_overrides,
            enable_application_ui_links=True,
        )
        operator.execute(self.mock_context)
        mock_conn.start_job_run.assert_called_once()

        mock_logs_link.assert_called_once_with(
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            aws_partition=mock.ANY,
            conn_id=mock.ANY,
            application_id=application_id,
            job_run_id=job_run_id,
        )
        mock_dashboard_link.assert_called_with(
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            aws_partition=mock.ANY,
            conn_id=mock.ANY,
            application_id=application_id,
            job_run_id=job_run_id,
        )
        mock_cloudwatch_link.assert_not_called()
        mock_s3_logs_link.assert_called_once_with(
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            aws_partition=mock.ANY,
            log_uri=s3_logs_location,
            application_id=application_id,
            job_run_id=job_run_id,
        )

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessCloudWatchLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessDashboardLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessLogsLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessS3LogsLink.persist")
    def test_links_spark_without_applicationui_enabled(
        self,
        mock_s3_logs_link,
        mock_logs_link,
        mock_dashboard_link,
        mock_cloudwatch_link,
        mock_conn,
        mock_get_waiter,
    ):
        mock_get_waiter.wait.return_value = True
        mock_conn.get_application.return_value = {"application": {"state": "STARTED"}}
        mock_conn.start_job_run.return_value = {
            "jobRunId": job_run_id,
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=spark_job_driver,
            configuration_overrides=s3_configuration_overrides,
            enable_application_ui_links=False,
        )
        operator.execute(self.mock_context)
        mock_conn.start_job_run.assert_called_once()

        mock_logs_link.assert_not_called()
        mock_dashboard_link.assert_not_called()
        mock_cloudwatch_link.assert_not_called()
        mock_s3_logs_link.assert_called_once_with(
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            aws_partition=mock.ANY,
            log_uri=s3_logs_location,
            application_id=application_id,
            job_run_id=job_run_id,
        )

    def test_template_fields(self):
        operator = EmrServerlessStartJobOperator(
            task_id=task_id,
            client_request_token=client_request_token,
            application_id=application_id,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
        )

        template_fields = list(operator.template_fields) + list(operator.template_fields_renderers.keys())

        class_fields = operator.__dict__

        missing_fields = [field for field in template_fields if field not in class_fields]

        assert not missing_fields, f"Templated fields are not available {missing_fields}"


class TestEmrServerlessDeleteOperator:
    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_delete_application_with_wait_for_completion_successfully(self, mock_conn, mock_get_waiter):
        mock_get_waiter().wait.return_value = True
        mock_conn.stop_application.return_value = {}
        mock_conn.delete_application.return_value = {"ResponseMetadata": {"HTTPStatusCode": 200}}

        operator = EmrServerlessDeleteApplicationOperator(
            task_id=task_id, application_id=application_id_delete_operator
        )

        operator.execute(None)

        assert operator.wait_for_completion is True
        assert mock_get_waiter().wait.call_count == 2
        mock_conn.stop_application.assert_called_once()
        mock_conn.delete_application.assert_called_once_with(applicationId=application_id_delete_operator)

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_delete_application_without_wait_for_completion_successfully(self, mock_conn, mock_get_waiter):
        mock_get_waiter().wait.return_value = True
        mock_conn.stop_application.return_value = {}
        mock_conn.delete_application.return_value = {"ResponseMetadata": {"HTTPStatusCode": 200}}

        operator = EmrServerlessDeleteApplicationOperator(
            task_id=task_id,
            application_id=application_id_delete_operator,
            wait_for_completion=False,
        )

        operator.execute(None)

        mock_get_waiter().wait.assert_called_once()
        mock_conn.stop_application.assert_called_once()
        mock_conn.delete_application.assert_called_once_with(applicationId=application_id_delete_operator)

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_delete_application_failed_deletion(self, mock_conn, mock_get_waiter):
        mock_get_waiter().wait.return_value = True
        mock_conn.stop_application.return_value = {}
        mock_conn.delete_application.return_value = {"ResponseMetadata": {"HTTPStatusCode": 400}}

        operator = EmrServerlessDeleteApplicationOperator(
            task_id=task_id, application_id=application_id_delete_operator
        )
        with pytest.raises(AirflowException) as ex_message:
            operator.execute(None)

        assert "Application deletion failed:" in str(ex_message.value)

        mock_get_waiter().wait.assert_called_once()
        mock_conn.stop_application.assert_called_once()
        mock_conn.delete_application.assert_called_once_with(applicationId=application_id_delete_operator)

    @pytest.mark.parametrize(
        ("waiter_delay", "waiter_max_attempts", "expected"),
        [
            (NOTSET, NOTSET, [60, 25]),
            (30, 10, [30, 10]),
        ],
    )
    def test_delete_application_waiter_params(
        self,
        waiter_delay,
        waiter_max_attempts,
        expected,
    ):
        operator = EmrServerlessDeleteApplicationOperator(
            task_id=task_id,
            application_id=application_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
        )
        assert operator.wait_for_completion is True
        assert operator.waiter_delay == expected[0]
        assert operator.waiter_max_attempts == expected[1]

    @mock.patch.object(EmrServerlessHook, "conn")
    def test_delete_application_deferrable(self, mock_conn):
        mock_conn.delete_application.return_value = {"ResponseMetadata": {"HTTPStatusCode": 200}}

        operator = EmrServerlessDeleteApplicationOperator(
            task_id=task_id,
            application_id=application_id,
            deferrable=True,
        )
        with pytest.raises(TaskDeferred):
            operator.execute(None)

    def test_template_fields(self):
        operator = EmrServerlessDeleteApplicationOperator(
            task_id=task_id, application_id=application_id_delete_operator
        )

        template_fields = list(operator.template_fields) + list(operator.template_fields_renderers.keys())

        class_fields = operator.__dict__

        missing_fields = [field for field in template_fields if field not in class_fields]

        assert not missing_fields, f"Templated fields are not available {missing_fields}"


class TestEmrServerlessStopOperator:
    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_stop(self, mock_conn: MagicMock, mock_get_waiter: MagicMock):
        mock_get_waiter().wait.return_value = True
        operator = EmrServerlessStopApplicationOperator(task_id=task_id, application_id="test")

        operator.execute({})

        mock_get_waiter().wait.assert_called_once()
        mock_conn.stop_application.assert_called_once()

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    def test_stop_no_wait(self, mock_conn: MagicMock, mock_get_waiter: MagicMock):
        operator = EmrServerlessStopApplicationOperator(
            task_id=task_id, application_id="test", wait_for_completion=False
        )

        operator.execute({})

        mock_get_waiter().wait.assert_not_called()
        mock_conn.stop_application.assert_called_once()

    @mock.patch.object(EmrServerlessHook, "get_waiter")
    @mock.patch.object(EmrServerlessHook, "conn")
    @mock.patch.object(EmrServerlessHook, "cancel_running_jobs")
    def test_force_stop(self, mock_cancel_running_jobs, mock_conn, mock_get_waiter):
        mock_cancel_running_jobs.return_value = 0
        mock_conn.stop_application.return_value = {}
        mock_get_waiter().wait.return_value = True

        operator = EmrServerlessStopApplicationOperator(
            task_id=task_id, application_id="test", force_stop=True
        )

        operator.execute({})

        mock_cancel_running_jobs.assert_called_once()
        mock_conn.stop_application.assert_called_once()
        mock_get_waiter().wait.assert_called_once()

    @mock.patch.object(EmrServerlessHook, "cancel_running_jobs")
    def test_stop_application_deferrable_with_force_stop(self, mock_cancel_running_jobs, caplog):
        mock_cancel_running_jobs.return_value = 2
        operator = EmrServerlessStopApplicationOperator(
            task_id=task_id, application_id="test", deferrable=True, force_stop=True
        )
        with pytest.raises(TaskDeferred):
            operator.execute({})
        assert "now waiting for the 2 cancelled job(s) to terminate" in caplog.messages

    @mock.patch.object(EmrServerlessHook, "conn")
    @mock.patch.object(EmrServerlessHook, "cancel_running_jobs")
    def test_stop_application_deferrable_without_force_stop(
        self, mock_cancel_running_jobs, mock_conn, caplog
    ):
        mock_conn.stop_application.return_value = {}
        mock_cancel_running_jobs.return_value = 0
        operator = EmrServerlessStopApplicationOperator(
            task_id=task_id, application_id="test", deferrable=True, force_stop=True
        )
        with pytest.raises(TaskDeferred):
            operator.execute({})

        assert "no running jobs found with application ID test" in caplog.messages

    def test_template_fields(self):
        operator = EmrServerlessStopApplicationOperator(
            task_id=task_id, application_id="test", deferrable=True, force_stop=True
        )

        validate_template_fields(operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.hooks.chime import ChimeWebhookHook
from airflow.providers.common.compat.notifier import BaseNotifier

if TYPE_CHECKING:
    from airflow.sdk import Context


class ChimeNotifier(BaseNotifier):
    """
    Chime notifier to send messages to a chime room via callbacks.

    :param chime_conn_id: The chime connection to use with Endpoint as "https://hooks.chime.aws" and
        the webhook token in the form of ```{webhook.id}?token{webhook.token}```
    :param message: The message to send to the chime room associated with the webhook.

    """

    template_fields = ("message",)

    def __init__(self, *, chime_conn_id: str, message: str = "This is the default chime notifier message"):
        super().__init__()
        self.chime_conn_id = chime_conn_id
        self.message = message

    @cached_property
    def hook(self):
        """To reduce overhead cache the hook for the notifier."""
        return ChimeWebhookHook(chime_conn_id=self.chime_conn_id)

    def notify(self, context: Context) -> None:
        """Send a message to a Chime Chat Room."""
        self.hook.send_message(message=self.message)


send_chime_notification = ChimeNotifier



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.ec2 import EC2Hook
from airflow.providers.amazon.aws.links.ec2 import (
    EC2InstanceDashboardLink,
    EC2InstanceLink,
)
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException

if TYPE_CHECKING:
    from airflow.sdk import Context


class EC2StartInstanceOperator(AwsBaseOperator[EC2Hook]):
    """
    Start AWS EC2 instance using boto3.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:EC2StartInstanceOperator`

    :param instance_id: id of the AWS EC2 instance
        :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param check_interval: time in seconds that the job should wait in
        between each instance state checks until operation is completed
    """

    aws_hook_class = EC2Hook
    operator_extra_links = (EC2InstanceLink(),)
    template_fields: Sequence[str] = aws_template_fields("instance_id", "region_name")
    ui_color = "#eeaa11"
    ui_fgcolor = "#ffffff"

    def __init__(
        self,
        *,
        instance_id: str,
        check_interval: float = 15,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.instance_id = instance_id
        self.check_interval = check_interval

    def execute(self, context: Context):
        self.log.info("Starting EC2 instance %s", self.instance_id)
        instance = self.hook.get_instance(instance_id=self.instance_id)
        instance.start()
        EC2InstanceLink.persist(
            context=context,
            operator=self,
            aws_partition=self.hook.conn_partition,
            instance_id=self.instance_id,
            region_name=self.hook.conn_region_name,
        )
        self.hook.wait_for_state(
            instance_id=self.instance_id,
            target_state="running",
            check_interval=self.check_interval,
        )


class EC2StopInstanceOperator(AwsBaseOperator[EC2Hook]):
    """
    Stop AWS EC2 instance using boto3.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:EC2StopInstanceOperator`

    :param instance_id: id of the AWS EC2 instance
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param check_interval: time in seconds that the job should wait in
        between each instance state checks until operation is completed
    """

    aws_hook_class = EC2Hook
    operator_extra_links = (EC2InstanceLink(),)
    template_fields: Sequence[str] = aws_template_fields("instance_id", "region_name")
    ui_color = "#eeaa11"
    ui_fgcolor = "#ffffff"

    def __init__(
        self,
        *,
        instance_id: str,
        check_interval: float = 15,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.instance_id = instance_id
        self.check_interval = check_interval

    def execute(self, context: Context):
        self.log.info("Stopping EC2 instance %s", self.instance_id)
        instance = self.hook.get_instance(instance_id=self.instance_id)
        EC2InstanceLink.persist(
            context=context,
            operator=self,
            aws_partition=self.hook.conn_partition,
            instance_id=self.instance_id,
            region_name=self.hook.conn_region_name,
        )
        instance.stop()

        self.hook.wait_for_state(
            instance_id=self.instance_id,
            target_state="stopped",
            check_interval=self.check_interval,
        )


class EC2CreateInstanceOperator(AwsBaseOperator[EC2Hook]):
    """
    Create and start a specified number of EC2 Instances using boto3.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:EC2CreateInstanceOperator`

    :param image_id: ID of the AMI used to create the instance.
    :param max_count: Maximum number of instances to launch. Defaults to 1.
    :param min_count: Minimum number of instances to launch. Defaults to 1.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param poll_interval: Number of seconds to wait before attempting to
        check state of instance. Only used if wait_for_completion is True. Default is 20.
    :param max_attempts: Maximum number of attempts when checking state of instance.
        Only used if wait_for_completion is True. Default is 20.
    :param config: Dictionary for arbitrary parameters to the boto3 run_instances call.
    :param wait_for_completion: If True, the operator will wait for the instance to be
        in the `running` state before returning.
    """

    aws_hook_class = EC2Hook

    operator_extra_links = (EC2InstanceDashboardLink(),)
    template_fields: Sequence[str] = aws_template_fields(
        "image_id",
        "max_count",
        "min_count",
        "aws_conn_id",
        "region_name",
        "config",
        "wait_for_completion",
    )

    def __init__(
        self,
        image_id: str,
        max_count: int = 1,
        min_count: int = 1,
        poll_interval: int = 20,
        max_attempts: int = 20,
        config: dict | None = None,
        wait_for_completion: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.image_id = image_id
        self.max_count = max_count
        self.min_count = min_count
        self.poll_interval = poll_interval
        self.max_attempts = max_attempts
        self.config = config or {}
        self.wait_for_completion = wait_for_completion

    @property
    def _hook_parameters(self) -> dict[str, Any]:
        return {**super()._hook_parameters, "api_type": "client_type"}

    def execute(self, context: Context):
        instances = self.hook.conn.run_instances(
            ImageId=self.image_id,
            MinCount=self.min_count,
            MaxCount=self.max_count,
            **self.config,
        )["Instances"]
        try:
            instance_ids = self._on_kill_instance_ids = [instance["InstanceId"] for instance in instances]
            # Console link is for EC2 dashboard list, not individual instances when more than 1 instance

            EC2InstanceDashboardLink.persist(
                context=context,
                operator=self,
                region_name=self.hook.conn_region_name,
                aws_partition=self.hook.conn_partition,
                instance_ids=EC2InstanceDashboardLink.format_instance_id_filter(instance_ids),
            )
            for instance_id in instance_ids:
                self.log.info("Created EC2 instance %s", instance_id)

                if self.wait_for_completion:
                    self.hook.get_waiter("instance_running").wait(
                        InstanceIds=[instance_id],
                        WaiterConfig={
                            "Delay": self.poll_interval,
                            "MaxAttempts": self.max_attempts,
                        },
                    )

            # leave "_on_kill_instance_ids" in place for finishing post-processing
            return instance_ids

        # Best-effort cleanup when post-creation steps fail (e.g. IAM/permission errors).
        except Exception:
            self.log.exception(
                "Exception after EC2 instance creation; attempting cleanup for instances %s",
                instance_ids,
            )
            try:
                self.hook.terminate_instances(instance_ids=instance_ids)
            except Exception:
                self.log.exception(
                    "Failed to cleanup EC2 instances %s after task failure",
                    instance_ids,
                )
            raise

    def on_kill(self) -> None:
        instance_ids = getattr(self, "_on_kill_instance_ids", [])

        if instance_ids:
            self.log.info("on_kill: Terminating instance/s %s", ", ".join(instance_ids))
            """ ec2_hook = EC2Hook(
                aws_conn_id=self.aws_conn_id,
                region_name=self.region_name,
                api_type="client_type",
            ) """
            self.hook.terminate_instances(instance_ids=instance_ids)
        super().on_kill()


class EC2TerminateInstanceOperator(AwsBaseOperator[EC2Hook]):
    """
    Terminate EC2 Instances using boto3.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:EC2TerminateInstanceOperator`

    :param instance_id: ID of the instance to be terminated.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param poll_interval: Number of seconds to wait before attempting to
        check state of instance. Only used if wait_for_completion is True. Default is 20.
    :param max_attempts: Maximum number of attempts when checking state of instance.
        Only used if wait_for_completion is True. Default is 20.
    :param wait_for_completion: If True, the operator will wait for the instance to be
        in the `terminated` state before returning.
    """

    aws_hook_class = EC2Hook
    template_fields: Sequence[str] = aws_template_fields(
        "instance_ids", "region_name", "aws_conn_id", "wait_for_completion"
    )

    def __init__(
        self,
        instance_ids: str | list[str],
        poll_interval: int = 20,
        max_attempts: int = 20,
        wait_for_completion: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.instance_ids = instance_ids
        self.poll_interval = poll_interval
        self.max_attempts = max_attempts
        self.wait_for_completion = wait_for_completion

    @property
    def _hook_parameters(self) -> dict[str, Any]:
        return {**super()._hook_parameters, "api_type": "client_type"}

    def execute(self, context: Context):
        if isinstance(self.instance_ids, str):
            self.instance_ids = [self.instance_ids]
        self.hook.conn.terminate_instances(InstanceIds=self.instance_ids)

        for instance_id in self.instance_ids:
            self.log.info("Terminating EC2 instance %s", instance_id)
            if self.wait_for_completion:
                self.hook.get_waiter("instance_terminated").wait(
                    InstanceIds=[instance_id],
                    WaiterConfig={
                        "Delay": self.poll_interval,
                        "MaxAttempts": self.max_attempts,
                    },
                )


class EC2RebootInstanceOperator(AwsBaseOperator[EC2Hook]):
    """
    Reboot Amazon EC2 instances.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:EC2RebootInstanceOperator`

    :param instance_ids: ID of the instance(s) to be rebooted.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param poll_interval: Number of seconds to wait before attempting to
        check state of instance. Only used if wait_for_completion is True. Default is 20.
    :param max_attempts: Maximum number of attempts when checking state of instance.
        Only used if wait_for_completion is True. Default is 20.
    :param wait_for_completion: If True, the operator will wait for the instance to be
        in the `running` state before returning.
    """

    aws_hook_class = EC2Hook
    operator_extra_links = (EC2InstanceDashboardLink(),)
    template_fields: Sequence[str] = aws_template_fields("instance_ids", "region_name")
    ui_color = "#eeaa11"
    ui_fgcolor = "#ffffff"

    def __init__(
        self,
        *,
        instance_ids: str | list[str],
        poll_interval: int = 20,
        max_attempts: int = 20,
        wait_for_completion: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.instance_ids = instance_ids
        self.poll_interval = poll_interval
        self.max_attempts = max_attempts
        self.wait_for_completion = wait_for_completion

    @property
    def _hook_parameters(self) -> dict[str, Any]:
        return {**super()._hook_parameters, "api_type": "client_type"}

    def execute(self, context: Context):
        if isinstance(self.instance_ids, str):
            self.instance_ids = [self.instance_ids]
        self.log.info("Rebooting EC2 instances %s", ", ".join(self.instance_ids))
        self.hook.conn.reboot_instances(InstanceIds=self.instance_ids)

        # Console link is for EC2 dashboard list, not individual instances
        EC2InstanceDashboardLink.persist(
            context=context,
            operator=self,
            region_name=self.hook.conn_region_name,
            aws_partition=self.hook.conn_partition,
            instance_ids=EC2InstanceDashboardLink.format_instance_id_filter(self.instance_ids),
        )
        if self.wait_for_completion:
            self.hook.get_waiter("instance_running").wait(
                InstanceIds=self.instance_ids,
                WaiterConfig={
                    "Delay": self.poll_interval,
                    "MaxAttempts": self.max_attempts,
                },
            )


class EC2HibernateInstanceOperator(AwsBaseOperator[EC2Hook]):
    """
    Hibernate Amazon EC2 instances.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:EC2HibernateInstanceOperator`

    :param instance_ids: ID of the instance(s) to be hibernated.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param poll_interval: Number of seconds to wait before attempting to
        check state of instance. Only used if wait_for_completion is True. Default is 20.
    :param max_attempts: Maximum number of attempts when checking state of instance.
        Only used if wait_for_completion is True. Default is 20.
    :param wait_for_completion: If True, the operator will wait for the instance to be
        in the `stopped` state before returning.
    """

    aws_hook_class = EC2Hook
    operator_extra_links = (EC2InstanceDashboardLink(),)
    template_fields: Sequence[str] = aws_template_fields("instance_ids", "region_name")
    ui_color = "#eeaa11"
    ui_fgcolor = "#ffffff"

    def __init__(
        self,
        *,
        instance_ids: str | list[str],
        poll_interval: int = 20,
        max_attempts: int = 20,
        wait_for_completion: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.instance_ids = instance_ids
        self.poll_interval = poll_interval
        self.max_attempts = max_attempts
        self.wait_for_completion = wait_for_completion

    @property
    def _hook_parameters(self) -> dict[str, Any]:
        return {**super()._hook_parameters, "api_type": "client_type"}

    def execute(self, context: Context):
        if isinstance(self.instance_ids, str):
            self.instance_ids = [self.instance_ids]
        self.log.info("Hibernating EC2 instances %s", ", ".join(self.instance_ids))
        instances = self.hook.get_instances(instance_ids=self.instance_ids)

        # Console link is for EC2 dashboard list, not individual instances
        EC2InstanceDashboardLink.persist(
            context=context,
            operator=self,
            region_name=self.hook.conn_region_name,
            aws_partition=self.hook.conn_partition,
            instance_ids=EC2InstanceDashboardLink.format_instance_id_filter(self.instance_ids),
        )

        for instance in instances:
            hibernation_options = instance.get("HibernationOptions")
            if not hibernation_options or not hibernation_options["Configured"]:
                raise AirflowException(f"Instance {instance['InstanceId']} is not configured for hibernation")

        self.hook.conn.stop_instances(InstanceIds=self.instance_ids, Hibernate=True)

        if self.wait_for_completion:
            self.hook.get_waiter("instance_stopped").wait(
                InstanceIds=self.instance_ids,
                WaiterConfig={
                    "Delay": self.poll_interval,
                    "MaxAttempts": self.max_attempts,
                },
            )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.amazon.aws.hooks.quicksight import QuickSightHook
from airflow.providers.amazon.aws.operators.quicksight import QuickSightCreateIngestionOperator

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

DATA_SET_ID = "DemoDataSet"
INGESTION_ID = "DemoDataSet_Ingestion"
AWS_ACCOUNT_ID = "123456789012"
INGESTION_TYPE = "FULL_REFRESH"

MOCK_RESPONSE = {
    "Status": 201,
    "Arn": "arn:aws:quicksight:us-east-1:123456789012:dataset/DemoDataSet/ingestion/DemoDataSet_Ingestion",
    "IngestionId": "DemoDataSet_Ingestion",
    "IngestionStatus": "INITIALIZED",
    "RequestId": "fc1f7eea-1327-41d6-9af7-c12f097ed343",
}


class TestQuickSightCreateIngestionOperator:
    def setup_method(self):
        self.default_op_kwargs = {
            "task_id": "quicksight_create",
            "aws_conn_id": None,
            "data_set_id": DATA_SET_ID,
            "ingestion_id": INGESTION_ID,
        }

    def test_init(self):
        self.default_op_kwargs.pop("aws_conn_id", None)

        op = QuickSightCreateIngestionOperator(
            **self.default_op_kwargs,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="cn-north-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.client_type == "quicksight"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "cn-north-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = QuickSightCreateIngestionOperator(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(QuickSightHook, "create_ingestion", return_value=MOCK_RESPONSE)
    def test_execute(self, mock_create_ingestion):
        QuickSightCreateIngestionOperator(**self.default_op_kwargs).execute({})
        mock_create_ingestion.assert_called_once_with(
            data_set_id=DATA_SET_ID,
            ingestion_id=INGESTION_ID,
            ingestion_type="FULL_REFRESH",
            wait_for_completion=True,
            check_interval=30,
        )

    def test_template_fields(self):
        operator = QuickSightCreateIngestionOperator(**self.default_op_kwargs)
        validate_template_fields(operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains AWS MWAA hook."""

from __future__ import annotations

import warnings
from typing import Literal

import requests
from botocore.exceptions import ClientError

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook


class MwaaHook(AwsBaseHook):
    """
    Interact with AWS Managed Workflows for Apache Airflow.

    Provide thin wrapper around :external+boto3:py:class:`boto3.client("mwaa") <MWAA.Client>`

    If your IAM policy doesn't have `airflow:InvokeRestApi` permission, the hook will use a fallback method
    that uses the AWS credential to generate a local web login token for the Airflow Web UI and then directly
    make requests to the Airflow API. This fallback method can be set as the default (and only) method used by
    setting `generate_local_token` to True.  Learn more here:
    https://docs.aws.amazon.com/mwaa/latest/userguide/access-mwaa-apache-airflow-rest-api.html#granting-access-MWAA-Enhanced-REST-API

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs["client_type"] = "mwaa"
        super().__init__(*args, **kwargs)

    def invoke_rest_api(
        self,
        env_name: str,
        path: str,
        method: str,
        body: dict | None = None,
        query_params: dict | None = None,
        generate_local_token: bool = False,
        airflow_version: Literal[2, 3] | None = None,
    ) -> dict:
        """
        Invoke the REST API on the Airflow webserver with the specified inputs.

        .. seealso::
            - :external+boto3:py:meth:`MWAA.Client.invoke_rest_api`

        :param env_name: name of the MWAA environment
        :param path: Apache Airflow REST API endpoint path to be called
        :param method: HTTP method used for making Airflow REST API calls: 'GET'|'PUT'|'POST'|'PATCH'|'DELETE'
        :param body: Request body for the Apache Airflow REST API call
        :param query_params: Query parameters to be included in the Apache Airflow REST API call
        :param generate_local_token: If True, only the local web token method is used without trying boto's
            `invoke_rest_api` first. If False, the local web token method is used as a fallback after trying
            boto's `invoke_rest_api`
        :param airflow_version: The Airflow major version the MWAA environment runs.
            This parameter is only used if the local web token method is used to call Airflow API.
        """
        # Filter out keys with None values because Airflow REST API doesn't accept requests otherwise
        body = {k: v for k, v in body.items() if v is not None} if body else {}
        query_params = query_params or {}
        api_kwargs = {
            "Name": env_name,
            "Path": path,
            "Method": method,
            "Body": body,
            "QueryParameters": query_params,
        }

        if generate_local_token:
            return self._invoke_rest_api_using_local_session_token(airflow_version, **api_kwargs)

        try:
            response = self.conn.invoke_rest_api(**api_kwargs)
            # ResponseMetadata is removed because it contains data that is either very unlikely to be useful
            # in XComs and logs, or redundant given the data already included in the response
            response.pop("ResponseMetadata", None)
            return response

        except ClientError as e:
            if (
                e.response["Error"]["Code"] == "AccessDeniedException"
                and "Airflow role" in e.response["Error"]["Message"]
            ):
                self.log.info(
                    "Access Denied due to missing airflow:InvokeRestApi in IAM policy. Trying again by generating local token..."
                )
                return self._invoke_rest_api_using_local_session_token(airflow_version, **api_kwargs)
            to_log = e.response
            # ResponseMetadata is removed because it contains data that is either very unlikely to be
            # useful in XComs and logs, or redundant given the data already included in the response
            to_log.pop("ResponseMetadata", None)
            self.log.error(to_log)
            raise

    def _invoke_rest_api_using_local_session_token(
        self,
        airflow_version: Literal[2, 3] | None = None,
        **api_kwargs,
    ) -> dict:
        if not airflow_version:
            warnings.warn(
                "The parameter ``airflow_version`` in ``MwaaHook.invoke_rest_api`` is not "
                "specified and the local web token method is being used. "
                "The default Airflow version being used is 2 but this value will change in the future. "
                "To avoid any unexpected behavior, please explicitly specify the Airflow version.",
                FutureWarning,
                stacklevel=3,
            )
            airflow_version = 2

        try:
            session, hostname, login_response = self._get_session_conn(api_kwargs["Name"], airflow_version)

            headers = {}
            if airflow_version == 3:
                headers = {
                    "Authorization": f"Bearer {login_response.cookies['_token']}",
                    "Content-Type": "application/json",
                }

            api_version = "v1" if airflow_version == 2 else "v2"
            response = session.request(
                method=api_kwargs["Method"],
                url=f"https://{hostname}/api/{api_version}{api_kwargs['Path']}",
                headers=headers,
                params=api_kwargs["QueryParameters"],
                json=api_kwargs["Body"],
                timeout=10,
            )
            response.raise_for_status()

        except requests.HTTPError as e:
            self.log.error(e.response.json())
            raise

        return {
            "RestApiStatusCode": response.status_code,
            "RestApiResponse": response.json(),
        }

    # Based on: https://docs.aws.amazon.com/mwaa/latest/userguide/access-mwaa-apache-airflow-rest-api.html#create-web-server-session-token
    def _get_session_conn(self, env_name: str, airflow_version: Literal[2, 3]) -> tuple:
        create_token_response = self.conn.create_web_login_token(Name=env_name)
        web_server_hostname = create_token_response["WebServerHostname"]
        web_token = create_token_response["WebToken"]

        login_url = (
            f"https://{web_server_hostname}/aws_mwaa/login"
            if airflow_version == 2
            else f"https://{web_server_hostname}/pluginsv2/aws_mwaa/login"
        )
        login_payload = {"token": web_token}
        session = requests.Session()
        login_response = session.post(login_url, data=login_payload, timeout=10)
        login_response.raise_for_status()

        return session, web_server_hostname, login_response



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import boto3
import botocore
import pytest

from airflow.providers.amazon.aws.hooks.kinesis_analytics import KinesisAnalyticsV2Hook
from airflow.providers.amazon.aws.sensors.kinesis_analytics import (
    KinesisAnalyticsV2StartApplicationCompletedSensor,
    KinesisAnalyticsV2StopApplicationCompletedSensor,
)


class TestKinesisAnalyticsV2CustomWaiters:
    def test_service_waiters(self):
        assert "application_start_complete" in KinesisAnalyticsV2Hook().list_waiters()
        assert "application_stop_complete" in KinesisAnalyticsV2Hook().list_waiters()


class TestKinesisAnalyticsV2CustomWaitersBase:
    @pytest.fixture(autouse=True)
    def mock_conn(self, monkeypatch):
        self.client = boto3.client("kinesisanalyticsv2")
        monkeypatch.setattr(KinesisAnalyticsV2Hook, "conn", self.client)


class TestKinesisAnalyticsV2ApplicationStartWaiter(TestKinesisAnalyticsV2CustomWaitersBase):
    APPLICATION_NAME = "demo"
    WAITER_NAME = "application_start_complete"

    @pytest.fixture
    def mock_describe_application(self):
        with mock.patch.object(self.client, "describe_application") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", KinesisAnalyticsV2StartApplicationCompletedSensor.SUCCESS_STATES)
    def test_start_application_complete(self, state, mock_describe_application):
        mock_describe_application.return_value = {"ApplicationDetail": {"ApplicationStatus": state}}

        KinesisAnalyticsV2Hook().get_waiter(self.WAITER_NAME).wait(ApplicationName=self.APPLICATION_NAME)

    @pytest.mark.parametrize("state", KinesisAnalyticsV2StartApplicationCompletedSensor.FAILURE_STATES)
    def test_start_application_complete_failed(self, state, mock_describe_application):
        mock_describe_application.return_value = {"ApplicationDetail": {"ApplicationStatus": state}}
        with pytest.raises(botocore.exceptions.WaiterError):
            KinesisAnalyticsV2Hook().get_waiter(self.WAITER_NAME).wait(ApplicationName=self.APPLICATION_NAME)

    def test_start_application_complete_wait(self, mock_describe_application):
        wait = {"ApplicationDetail": {"ApplicationStatus": "STARTING"}}
        success = {"ApplicationDetail": {"ApplicationStatus": "RUNNING"}}

        mock_describe_application.side_effect = [wait, wait, success]

        KinesisAnalyticsV2Hook().get_waiter(self.WAITER_NAME).wait(
            ApplicationName=self.APPLICATION_NAME, WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )


class TestKinesisAnalyticsV2ApplicationStopWaiter(TestKinesisAnalyticsV2CustomWaitersBase):
    APPLICATION_NAME = "demo"
    WAITER_NAME = "application_stop_complete"

    @pytest.fixture
    def mock_describe_application(self):
        with mock.patch.object(self.client, "describe_application") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", KinesisAnalyticsV2StopApplicationCompletedSensor.SUCCESS_STATES)
    def test_stop_application_complete(self, state, mock_describe_application):
        mock_describe_application.return_value = {"ApplicationDetail": {"ApplicationStatus": state}}

        KinesisAnalyticsV2Hook().get_waiter(self.WAITER_NAME).wait(ApplicationName=self.APPLICATION_NAME)

    @pytest.mark.parametrize("state", KinesisAnalyticsV2StopApplicationCompletedSensor.FAILURE_STATES)
    def test_stop_application_complete_failed(self, state, mock_describe_application):
        mock_describe_application.return_value = {"ApplicationDetail": {"ApplicationStatus": state}}
        with pytest.raises(botocore.exceptions.WaiterError):
            KinesisAnalyticsV2Hook().get_waiter(self.WAITER_NAME).wait(ApplicationName=self.APPLICATION_NAME)

    def test_stop_application_complete_wait(self, mock_describe_application):
        wait = {"ApplicationDetail": {"ApplicationStatus": "STOPPING"}}
        success = {"ApplicationDetail": {"ApplicationStatus": "READY"}}

        mock_describe_application.side_effect = [wait, wait, success]

        KinesisAnalyticsV2Hook().get_waiter(self.WAITER_NAME).wait(
            ApplicationName=self.APPLICATION_NAME, WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar

from botocore.exceptions import ClientError

from airflow.providers.amazon.aws.hooks.kinesis_analytics import KinesisAnalyticsV2Hook
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.triggers.kinesis_analytics import (
    KinesisAnalyticsV2ApplicationOperationCompleteTrigger,
)
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException, conf

if TYPE_CHECKING:
    from airflow.sdk import Context


class KinesisAnalyticsV2CreateApplicationOperator(AwsBaseOperator[KinesisAnalyticsV2Hook]):
    """
    Creates an AWS Managed Service for Apache Flink application.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:KinesisAnalyticsV2CreateApplicationOperator`

    :param application_name: The name of application. (templated)
    :param runtime_environment: The runtime environment for the application. (templated)
    :param service_execution_role: The IAM role used by the application to access services. (templated)
    :param create_application_kwargs: Create application extra properties. (templated)
    :param application_description: A summary description of the application. (templated)

    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = KinesisAnalyticsV2Hook
    ui_color = "#44b5e2"

    template_fields: Sequence[str] = aws_template_fields(
        "application_name",
        "runtime_environment",
        "service_execution_role",
        "create_application_kwargs",
        "application_description",
    )
    template_fields_renderers: ClassVar[dict] = {
        "create_application_kwargs": "json",
    }

    def __init__(
        self,
        application_name: str,
        runtime_environment: str,
        service_execution_role: str,
        create_application_kwargs: dict[str, Any] | None = None,
        application_description: str = "Managed Service for Apache Flink application created from Airflow",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.application_name = application_name
        self.runtime_environment = runtime_environment
        self.service_execution_role = service_execution_role
        self.create_application_kwargs = create_application_kwargs or {}
        self.application_description = application_description

    def execute(self, context: Context) -> dict[str, str]:
        self.log.info("Creating AWS Managed Service for Apache Flink application %s.", self.application_name)
        try:
            response = self.hook.conn.create_application(
                ApplicationName=self.application_name,
                ApplicationDescription=self.application_description,
                RuntimeEnvironment=self.runtime_environment,
                ServiceExecutionRole=self.service_execution_role,
                **self.create_application_kwargs,
            )
        except ClientError as error:
            raise AirflowException(
                f"AWS Managed Service for Apache Flink application creation failed: {error.response['Error']['Message']}"
            )

        self.log.info(
            "AWS Managed Service for Apache Flink application created successfully %s.",
            self.application_name,
        )

        return {"ApplicationARN": response["ApplicationDetail"]["ApplicationARN"]}


class KinesisAnalyticsV2StartApplicationOperator(AwsBaseOperator[KinesisAnalyticsV2Hook]):
    """
    Starts an AWS Managed Service for Apache Flink application.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:KinesisAnalyticsV2StartApplicationOperator`

    :param application_name: The name of application. (templated)
    :param run_configuration: Application properties to start Apache Flink Job. (templated)

    :param wait_for_completion: Whether to wait for job to stop. (default: True)
    :param waiter_delay: Time in seconds to wait between status checks. (default: 60)
    :param waiter_max_attempts: Maximum number of attempts to check for job completion. (default: 20)
    :param deferrable: If True, the operator will wait asynchronously for the job to stop.
        This implies waiting for completion. This mode requires aiobotocore module to be installed.
        (default: False)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = KinesisAnalyticsV2Hook
    ui_color = "#44b5e2"

    template_fields: Sequence[str] = aws_template_fields(
        "application_name",
        "run_configuration",
    )
    template_fields_renderers: ClassVar[dict] = {
        "run_configuration": "json",
    }

    def __init__(
        self,
        application_name: str,
        run_configuration: dict[str, Any] | None = None,
        wait_for_completion: bool = True,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 20,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.application_name = application_name
        self.run_configuration = run_configuration or {}
        self.wait_for_completion = wait_for_completion
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts
        self.deferrable = deferrable

    def execute(self, context: Context) -> dict[str, Any]:
        msg = "AWS Managed Service for Apache Flink application"

        try:
            self.log.info("Starting %s %s.", msg, self.application_name)
            self.hook.conn.start_application(
                ApplicationName=self.application_name, RunConfiguration=self.run_configuration
            )
        except ClientError as error:
            raise AirflowException(
                f"Failed to start {msg} {self.application_name}: {error.response['Error']['Message']}"
            )

        describe_response = self.hook.conn.describe_application(ApplicationName=self.application_name)

        if self.deferrable:
            self.log.info("Deferring for %s to start: %s.", msg, self.application_name)
            self.defer(
                trigger=KinesisAnalyticsV2ApplicationOperationCompleteTrigger(
                    application_name=self.application_name,
                    waiter_name="application_start_complete",
                    aws_conn_id=self.aws_conn_id,
                    waiter_delay=self.waiter_delay,
                    waiter_max_attempts=self.waiter_max_attempts,
                    region_name=self.region_name,
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="execute_complete",
            )
        if self.wait_for_completion:
            self.log.info("Waiting for %s to start: %s.", msg, self.application_name)

            self.hook.get_waiter("application_start_complete").wait(
                ApplicationName=self.application_name,
                WaiterConfig={"Delay": self.waiter_delay, "MaxAttempts": self.waiter_max_attempts},
            )

        self.log.info("%s started successfully %s.", msg, self.application_name)

        return {"ApplicationARN": describe_response["ApplicationDetail"]["ApplicationARN"]}

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> dict[str, Any]:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException(
                "Error while starting AWS Managed Service for Apache Flink application: %s", validated_event
            )

        response = self.hook.conn.describe_application(
            ApplicationName=validated_event["application_name"],
        )

        self.log.info(
            "AWS Managed Service for Apache Flink application %s started successfully.",
            validated_event["application_name"],
        )

        return {"ApplicationARN": response["ApplicationDetail"]["ApplicationARN"]}


class KinesisAnalyticsV2StopApplicationOperator(AwsBaseOperator[KinesisAnalyticsV2Hook]):
    """
    Stop an AWS Managed Service for Apache Flink application.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:KinesisAnalyticsV2StopApplicationOperator`

    :param application_name: The name of your application. (templated)
    :param force: Set to true to force the application to stop. If you set Force to true, Managed Service for
        Apache Flink stops the application without taking a snapshot. (templated)

    :param wait_for_completion: Whether to wait for job to stop. (default: True)
    :param waiter_delay: Time in seconds to wait between status checks. (default: 60)
    :param waiter_max_attempts: Maximum number of attempts to check for job completion. (default: 20)
    :param deferrable: If True, the operator will wait asynchronously for the job to stop.
        This implies waiting for completion. This mode requires aiobotocore module to be installed.
        (default: False)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = KinesisAnalyticsV2Hook
    ui_color = "#44b5e2"

    template_fields: Sequence[str] = aws_template_fields(
        "application_name",
        "force",
    )

    def __init__(
        self,
        application_name: str,
        force: bool = False,
        wait_for_completion: bool = True,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 20,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.application_name = application_name
        self.force = force
        self.wait_for_completion = wait_for_completion
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts
        self.deferrable = deferrable

    def execute(self, context: Context) -> dict[str, Any]:
        msg = "AWS Managed Service for Apache Flink application"

        try:
            self.log.info("Stopping %s %s.", msg, self.application_name)

            self.hook.conn.stop_application(ApplicationName=self.application_name, Force=self.force)
        except ClientError as error:
            raise AirflowException(
                f"Failed to stop {msg} {self.application_name}: {error.response['Error']['Message']}"
            )

        describe_response = self.hook.conn.describe_application(ApplicationName=self.application_name)

        if self.deferrable:
            self.log.info("Deferring for %s to stop: %s.", msg, self.application_name)
            self.defer(
                trigger=KinesisAnalyticsV2ApplicationOperationCompleteTrigger(
                    application_name=self.application_name,
                    waiter_name="application_stop_complete",
                    aws_conn_id=self.aws_conn_id,
                    waiter_delay=self.waiter_delay,
                    waiter_max_attempts=self.waiter_max_attempts,
                    region_name=self.region_name,
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="execute_complete",
            )
        if self.wait_for_completion:
            self.log.info("Waiting for %s to stop: %s.", msg, self.application_name)

            self.hook.get_waiter("application_stop_complete").wait(
                ApplicationName=self.application_name,
                WaiterConfig={"Delay": self.waiter_delay, "MaxAttempts": self.waiter_max_attempts},
            )

        self.log.info("%s stopped successfully %s.", msg, self.application_name)

        return {"ApplicationARN": describe_response["ApplicationDetail"]["ApplicationARN"]}

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> dict[str, Any]:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException("Error while stopping AWS Managed Service for Apache Flink application")

        response = self.hook.conn.describe_application(
            ApplicationName=validated_event["application_name"],
        )

        self.log.info(
            "AWS Managed Service for Apache Flink application %s stopped successfully.",
            validated_event["application_name"],
        )

        return {"ApplicationARN": response["ApplicationDetail"]["ApplicationARN"]}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""This module contains the Amazon SageMaker Unified Studio Notebook job trigger."""

from __future__ import annotations

from airflow.triggers.base import BaseTrigger


class SageMakerNotebookJobTrigger(BaseTrigger):
    """
    Watches for a notebook job, triggers when it finishes.

    Examples:
     .. code-block:: python

        from airflow.providers.amazon.aws.triggers.sagemaker_unified_studio import SageMakerNotebookJobTrigger

        notebook_trigger = SageMakerNotebookJobTrigger(
            execution_id="notebook_job_1234",
            execution_name="notebook_task",
            waiter_delay=10,
            waiter_max_attempts=1440,
        )

    :param execution_id: A unique, meaningful id for the task.
    :param execution_name: A unique, meaningful name for the task.
    :param waiter_delay: Interval in seconds to check the notebook execution status.
    :param waiter_max_attempts: Number of attempts to wait before returning FAILED.
    """

    def __init__(self, execution_id, execution_name, waiter_delay, waiter_max_attempts, **kwargs):
        super().__init__(**kwargs)
        self.execution_id = execution_id
        self.execution_name = execution_name
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts

    def serialize(self):
        return (
            # dynamically generate the fully qualified name of the class
            self.__class__.__module__ + "." + self.__class__.__qualname__,
            {
                "execution_id": self.execution_id,
                "execution_name": self.execution_name,
                "poll_interval": self.poll_interval,
            },
        )

    async def run(self):
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from functools import cache
from pathlib import Path
from unittest.mock import Mock, patch

import boto3
import pytest

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if not AIRFLOW_V_3_0_PLUS:
    pytest.skip("AWS auth manager is only compatible with Airflow >= 3.0.0", allow_module_level=True)

from fastapi.testclient import TestClient
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser

from airflow.api_fastapi.app import AUTH_MANAGER_FASTAPI_APP_PREFIX, create_app
from airflow.api_fastapi.auth.managers.base_auth_manager import COOKIE_NAME_JWT_TOKEN

from system.amazon.aws.utils import set_env_id
from tests_common.test_utils.config import conf_vars

SAML_METADATA_URL = "/saml/metadata"
SAML_METADATA_PARSED = {
    "idp": {
        "entityId": "https://portal.sso.us-east-1.amazonaws.com/saml/assertion/<assertion>",
        "singleSignOnService": {
            "url": "https://portal.sso.us-east-1.amazonaws.com/saml/assertion/<assertion>",
            "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
        },
        "singleLogoutService": {
            "url": "https://portal.sso.us-east-1.amazonaws.com/saml/logout/<assertion>",
            "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
        },
        "x509cert": "<cert>",
    },
    "security": {"authnRequestsSigned": False},
    "sp": {"NameIDFormat": "urn:oasis:names:tc:SAML:2.0:nameid-format:transient"},
}

AVP_POLICY_ADMIN = """
permit (
    principal in Airflow::Group::"Admin",
    action,
    resource
);
"""


def create_avp_policy_store(env_id):
    description = f"Created by system test TestAwsAuthManager: {env_id}"
    client = boto3.client("verifiedpermissions")
    response = client.create_policy_store(
        validationSettings={"mode": "STRICT"},
        description=description,
    )
    policy_store_id = response["policyStoreId"]

    schema_path = (
        Path(__file__)
        .parents[5]
        .joinpath("src", "airflow", "providers", "amazon", "aws", "auth_manager", "avp", "schema.json")
        .resolve()
    )
    with open(schema_path) as schema_file:
        client.put_schema(
            policyStoreId=policy_store_id,
            definition={
                "cedarJson": schema_file.read(),
            },
        )

    client.create_policy(
        policyStoreId=policy_store_id,
        definition={
            "static": {"description": "Admin permissions", "statement": AVP_POLICY_ADMIN},
        },
    )

    return policy_store_id


@cache
def env_id():
    return set_env_id("test_aws_auth_manager")


@pytest.fixture
def region_name():
    return boto3.session.Session().region_name


@pytest.fixture
@cache
def avp_policy_store_id():
    return create_avp_policy_store(env_id())


@pytest.fixture
def base_app(region_name, avp_policy_store_id):
    with conf_vars(
        {
            (
                "core",
                "auth_manager",
            ): "airflow.providers.amazon.aws.auth_manager.aws_auth_manager.AwsAuthManager",
            ("aws_auth_manager", "conn_id"): "aws_default",
            ("aws_auth_manager", "region_name"): region_name,
            ("aws_auth_manager", "saml_metadata_url"): SAML_METADATA_URL,
            ("aws_auth_manager", "avp_policy_store_id"): avp_policy_store_id,
        }
    ):
        with (
            patch.object(OneLogin_Saml2_IdPMetadataParser, "parse_remote") as mock_parse_remote,
            patch(
                "airflow.providers.amazon.aws.auth_manager.routes.login._init_saml_auth"
            ) as mock_init_saml_auth,
        ):
            mock_parse_remote.return_value = SAML_METADATA_PARSED
            yield mock_init_saml_auth


@pytest.fixture
def client_no_permissions(base_app):
    auth = Mock()
    auth.is_authenticated.return_value = True
    auth.get_nameid.return_value = "user_no_permissions"
    auth.get_attributes.return_value = {
        "id": ["user_no_permissions"],
        "groups": [],
        "email": ["email"],
    }
    base_app.return_value = auth
    return TestClient(create_app())


@pytest.fixture
def client_admin_permissions(base_app):
    auth = Mock()
    auth.is_authenticated.return_value = True
    auth.get_nameid.return_value = "user_admin_permissions"
    auth.get_attributes.return_value = {
        "id": ["user_admin_permissions"],
        "groups": ["Admin"],
    }
    base_app.return_value = auth
    return TestClient(create_app())


@pytest.mark.system
class TestAwsAuthManager:
    """
    Run tests on Airflow using AWS auth manager with real credentials
    """

    @classmethod
    def teardown_class(cls):
        cls.delete_avp_policy_store()

    @classmethod
    def delete_avp_policy_store(cls):
        client = boto3.client("verifiedpermissions")

        paginator = client.get_paginator("list_policy_stores")
        pages = paginator.paginate()
        policy_store_ids = [
            store["policyStoreId"]
            for page in pages
            for store in page["policyStores"]
            if "description" in store
            and f"Created by system test TestAwsAuthManager: {env_id()}" in store["description"]
        ]

        for policy_store_id in policy_store_ids:
            client.delete_policy_store(policyStoreId=policy_store_id)

    def test_login_admin_redirect(self, client_admin_permissions):
        response = client_admin_permissions.post(
            AUTH_MANAGER_FASTAPI_APP_PREFIX + "/login_callback",
            follow_redirects=False,
            data={"RelayState": "login-redirect"},
        )
        token = response.cookies.get(COOKIE_NAME_JWT_TOKEN)
        assert response.status_code == 303
        assert "location" in response.headers
        assert response.headers["location"] == "/"
        assert token is not None

    def test_login_admin_token(self, client_admin_permissions):
        response = client_admin_permissions.post(
            AUTH_MANAGER_FASTAPI_APP_PREFIX + "/login_callback",
            follow_redirects=False,
            data={"RelayState": "login-token"},
        )
        assert response.status_code == 200
        assert response.json()["access_token"]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.emr import EmrModifyClusterOperator
from airflow.providers.common.compat.sdk import AirflowException

from tests_common.test_utils.compat import timezone
from unit.amazon.aws.utils.test_template_fields import validate_template_fields

DEFAULT_DATE = timezone.datetime(2017, 1, 1)
MODIFY_CLUSTER_SUCCESS_RETURN = {"ResponseMetadata": {"HTTPStatusCode": 200}, "StepConcurrencyLevel": 1}
MODIFY_CLUSTER_ERROR_RETURN = {"ResponseMetadata": {"HTTPStatusCode": 400}}


@pytest.fixture
def mocked_hook_client():
    with patch("airflow.providers.amazon.aws.hooks.emr.EmrHook.conn") as m:
        yield m


class TestEmrModifyClusterOperator:
    def setup_method(self):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}

        self.mock_context = MagicMock()
        self.operator = EmrModifyClusterOperator(
            task_id="test_task",
            cluster_id="j-8989898989",
            step_concurrency_level=1,
            aws_conn_id="aws_default",
            dag=DAG("test_dag_id", schedule=None, default_args=args),
        )

    def test_init(self):
        assert self.operator.cluster_id == "j-8989898989"
        assert self.operator.step_concurrency_level == 1
        assert self.operator.aws_conn_id == "aws_default"

    def test_execute_returns_step_concurrency(self, mocked_hook_client):
        mocked_hook_client.modify_cluster.return_value = MODIFY_CLUSTER_SUCCESS_RETURN

        assert self.operator.execute(self.mock_context) == 1

    def test_execute_returns_error(self, mocked_hook_client):
        mocked_hook_client.modify_cluster.return_value = MODIFY_CLUSTER_ERROR_RETURN

        with pytest.raises(AirflowException, match="Modify cluster failed"):
            self.operator.execute(self.mock_context)

    def test_template_fields(self):
        validate_template_fields(self.operator)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains the Amazon DynamoDB Hook."""

from __future__ import annotations

from collections.abc import Iterable
from functools import cached_property
from typing import TYPE_CHECKING

from botocore.exceptions import ClientError

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.providers.common.compat.sdk import AirflowException

if TYPE_CHECKING:
    from botocore.client import BaseClient


class DynamoDBHook(AwsBaseHook):
    """
    Interact with Amazon DynamoDB.

    Provide thick wrapper around
    :external+boto3:py:class:`boto3.resource("dynamodb") <DynamoDB.ServiceResource>`.

    :param table_keys: partition key and sort key
    :param table_name: target DynamoDB table

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    def __init__(
        self, *args, table_keys: list | None = None, table_name: str | None = None, **kwargs
    ) -> None:
        self.table_keys = table_keys
        self.table_name = table_name
        kwargs["resource_type"] = "dynamodb"
        super().__init__(*args, **kwargs)

    @cached_property
    def client(self) -> BaseClient:
        """Return boto3 client."""
        return self.get_conn().meta.client

    def write_batch_data(self, items: Iterable) -> bool:
        """
        Write batch items to DynamoDB table with provisioned throughout capacity.

        .. seealso::
            - :external+boto3:py:meth:`DynamoDB.ServiceResource.Table`
            - :external+boto3:py:meth:`DynamoDB.Table.batch_writer`
            - :external+boto3:py:meth:`DynamoDB.Table.put_item`

        :param items: list of DynamoDB items.
        """
        try:
            table = self.get_conn().Table(self.table_name)

            with table.batch_writer(overwrite_by_pkeys=self.table_keys) as batch:
                for item in items:
                    batch.put_item(Item=item)
            return True
        except Exception as general_error:
            raise AirflowException(f"Failed to insert items in dynamodb, error: {general_error}")

    def get_import_status(self, import_arn: str) -> tuple[str, str | None, str | None]:
        """
        Get import status from Dynamodb.

        :param import_arn: The Amazon Resource Name (ARN) for the import.
        :return: Import status, Error code and Error message
        """
        self.log.info("Poking for Dynamodb import %s", import_arn)

        try:
            describe_import = self.client.describe_import(ImportArn=import_arn)
            status = describe_import["ImportTableDescription"]["ImportStatus"]
            error_code = describe_import["ImportTableDescription"].get("FailureCode")
            error_msg = describe_import["ImportTableDescription"].get("FailureMessage")
            return status, error_code, error_msg
        except ClientError as e:
            error_code = e.response.get("Error", {}).get("Code")
            if error_code == "ImportNotFoundException":
                raise AirflowException("S3 import into Dynamodb job not found.")
            raise e
